@@ -10,14 +10,26 @@
     identical(x = x$get_obj_name(), y = y$get_obj_name())
 }
 
+## h5group check existing ======================================================
+
+.h5group_exist <- function(h5group, name, ...) {
+  if (identical(x = name, y = "/")) {
+    return(TRUE)
+  }
+  return(tryCatch(
+    expr = h5group$exists(name = name, ...),
+    error = function(e) FALSE
+  ))
+}
+
 ## Exclude hdf5 link names =====================================================
+
 .exclude_h5_links <- function(all_links, exclude) {
   exclude <- vapply(
     X = exclude, 
     FUN = h5AbsLinkName, 
     FUN.VALUE = character(length = 1L)
   )
-  exclude <- gsub(pattern = "\\/*$", replacement = "", x = exclude)
   keep_links0 <- setdiff(all_links, exclude)
   keep_links <- keep_links0
   while (any(!keep_links %in% "/")) {
@@ -50,9 +62,8 @@
   return(attr_data)
 }
 
-#' @importFrom hdf5r h5garbage_collect
 .h5attr_overwrite <- function(h5obj, which, overwrite = TRUE) {
-  if (!h5obj$attr_exists_by_name(attr_name = which, obj_name = ".")) {
+  if (!h5obj$attr_exists(attr_name = which)) {
     return(invisible(x = NULL))
   }
   if (!overwrite) {
@@ -62,12 +73,11 @@
       "\n  Object: ", h5obj$get_obj_name(),
       "\n  Attribute: ", which, 
       "\nSet 'overwrite = TRUE' to overwrite it.",
-      immediate. = TRUE
+      immediate. = TRUE, call. = FALSE
     )
     return(invisible(x = NULL))
   }
-  h5obj$attr_delete_by_name(attr_name = which, obj_name = ".")
-  h5garbage_collect()
+  h5obj$attr_delete(attr_name = which)
   return(invisible(x = NULL))
 }
 
@@ -88,32 +98,30 @@
     h5space <- H5S$new(dims = 0, maxdims = 0)
     on.exit(expr = h5space$close(), add = TRUE)
     h5a <- h5obj$create_attr(
-      attr_name = "column-order", 
+      attr_name = which, 
       dtype = h5GuessDtype(x = robj),
       space = h5space
     )
+    on.exit(expr = h5a$close(), add = TRUE)
     return(invisible(x = NULL))
   }
   if (check.scalar && is_scalar_atomic(x = robj)) {
     h5space <- H5S$new(type = "scalar")
-    on.exit(expr = h5space$close())
+    on.exit(expr = h5space$close(), add = TRUE)
   }
-  attr_obj <- h5obj$create_attr_by_name(
+  h5a <- h5obj$create_attr(
     attr_name = which, 
-    obj_name = ".", 
     robj = robj,
     space = h5space,
     dtype = h5GuessDtype(x = robj, stype = stype)
   )
-  on.exit(expr = attr_obj$close(), add = TRUE)
+  on.exit(expr = h5a$close(), add = TRUE)
   return(invisible(x = NULL))
 }
 
-#' @importFrom hdf5r h5garbage_collect
 .h5attr_delete <- function(h5obj, which) {
-  if (h5obj$attr_exists_by_name(attr_name = which, obj_name = ".")) {
-    h5obj$attr_delete_by_name(attr_name = which, obj_name = ".")
-    h5garbage_collect()
+  if (h5obj$attr_exists(attr_name = which)) {
+    h5obj$attr_delete(attr_name = which)
   }
   return(invisible(x = NULL))
 }
@@ -125,7 +133,7 @@
     to.which,
     overwrite = TRUE
 ) {
-  if (!from.h5obj$attr_exists_by_name(attr_name = from.which, obj_name = ".")) {
+  if (!from.h5obj$attr_exists(attr_name = from.which)) {
     stop(
       "\nSource attribute doesn't exist:",
       "\n  File: ", from.h5obj$get_filename(),
@@ -137,11 +145,11 @@
     identical(x = from.which, y = to.which)
   if (is.identical) {
     warning(
-      "Source attribute is identical to the destination, skip copying: ",
+      "No copying. Source attribute is identical to the destination: ",
       "\n  File: ", from.h5obj$get_filename(),
       "\n  Object: ", from.h5obj$get_obj_name(),
       "\n  Attribute: ", from.which,
-      immediate. = TRUE
+      immediate. = TRUE, call. = FALSE
     )
     return(invisible(x = NULL))
   }
@@ -156,9 +164,8 @@
     expr = from.h5a$read(),
     error = function(e) NULL
   )
-  h5a <- to.h5obj$create_attr_by_name(
+  h5a <- to.h5obj$create_attr(
     attr_name = to.which, 
-    obj_name = ".", 
     robj = robj,
     space = h5space,
     dtype = h5dtype
@@ -177,7 +184,7 @@
 ) {
   from.h5obj <- h5Open(x = from.h5fh, name = from.name)
   if (!identical(x = from.h5obj, y = from.h5fh)) {
-    on.exit(expr = from.h5obj$close())
+    on.exit(expr = from.h5obj$close(), add = TRUE)
   }
   to.h5obj <- h5Open(x = to.h5fh, name = to.name)
   if (!identical(x = to.h5obj, y = to.h5fh)) {
@@ -196,7 +203,7 @@
   return(invisible(x = NULL))
 }
 
-## H5 copy or delete ===========================================================
+## H5 copy =====================================================================
 
 .h5copy_same_file <- function(
     h5.file, 
@@ -208,8 +215,8 @@
 ) {
   if (identical(x = from.name, y = to.name)) {
     warning(
-      "The source object is identical to the destination, skip copying.",
-      immediate. = TRUE
+      "No copying. The source object is identical to the destination. ",
+      immediate. = TRUE, call. = FALSE
     )
     return(invisible(x = NULL))
   }
@@ -223,7 +230,7 @@
       warning(
         "Destination object already exists. ",
         "Set 'overwrite = TRUE' to remove it.",
-        immediate. = TRUE
+        immediate. = TRUE, call. = FALSE
       )
       return(invisible(x = NULL))
     }
@@ -243,6 +250,7 @@
     dst_name = to.name, 
     ...
   )
+  h5fh$flush()
   .h5attr_copy_all(
     from.h5fh = h5fh,
     from.name = from.name,
@@ -262,23 +270,7 @@
     verbose = TRUE,
     ...
 ) {
-  if (!identical(x = to.name, y = "/")) {
-    to.h5fh <- h5TryOpen(filename = to.file, mode = "a")
-    if (h5Exists(x = to.h5fh, name = to.name)) {
-      if (!overwrite) {
-        warning(
-          "Destination object already exists. ",
-          "Set 'overwrite = TRUE' to remove it.",
-          immediate. = TRUE
-        )
-        return(invisible(x = NULL))
-      }
-      if (verbose) {
-        message("Destination object already exists, removing it.")
-      }
-      to.h5fh$link_delete(name = to.name)
-    }
-  } else {
+  if (identical(x = to.name, y = "/")) {
     if (file.exists(to.file) && !overwrite) {
       warning(
         "Destination file already exists. ",
@@ -295,8 +287,25 @@
       return(invisible(x = NULL))
     }
     to.h5fh <- h5TryOpen(filename = to.file, mode = "w")
+    on.exit(expr = to.h5fh$close(), add = TRUE)
+  } else {
+    to.h5fh <- h5TryOpen(filename = to.file, mode = "a")
+    on.exit(expr = to.h5fh$close(), add = TRUE)
+    if (h5Exists(x = to.h5fh, name = to.name)) {
+      if (!overwrite) {
+        warning(
+          "Destination object already exists. ",
+          "Set 'overwrite = TRUE' to remove it.",
+          immediate. = TRUE, call. = FALSE
+        )
+        return(invisible(x = NULL))
+      }
+      if (verbose) {
+        message("Destination object already exists, removing it.")
+      }
+      to.h5fh$link_delete(name = to.name)
+    }
   }
-  on.exit(expr = to.h5fh$close())
   h5fh <- h5TryOpen(filename = from.file, mode = "r")
   on.exit(expr = h5fh$close(), add = TRUE)
   if (!h5Exists(x = h5fh, name = from.name)) {
@@ -323,6 +332,8 @@
   return(invisible(x = NULL))
 }
 
+## h5 delete ===================================================================
+
 .h5delete <- function(h5obj, name, verbose = TRUE, ...) {
   if (verbose) {
     message(
@@ -333,7 +344,10 @@
     )
   }
   if (!h5Exists(x = h5obj, name = name)) {
-    warning("The H5 object to be deleted doesn't exists.", immediate. = TRUE)
+    warning(
+      "The H5 object to be deleted doesn't exists.", 
+      immediate. = TRUE, call. = FALSE
+    )
     return(invisible(x = NULL))
   }
   h5obj$link_delete(name = name, ...)
@@ -380,8 +394,8 @@
     if (show.warnings) {
       warning(
         "H5 group already exists: ", file_warning, group_warning,
-        "\n  Target group: ", path,
-        immediate. = TRUE
+        "\n  New sub-group: ", path,
+        immediate. = TRUE, call. = FALSE
       )
     }
   }
@@ -498,7 +512,7 @@
 ) {
   h5d <- h5Open(x = h5group, name = name)
   if (!identical(x = h5d, y = h5group)) {
-    on.exit(expr = h5d$close())
+    on.exit(expr = h5d$close(), add = TRUE)
   }
   if (!inherits(x = h5d, "H5D")) {
     stop("\n  '", h5d$get_obj_name(), "' is not an H5D")
@@ -523,7 +537,7 @@
 ) {
   h5d <- h5Open(x = h5group, name = name)
   if (!identical(x = h5d, y = h5group)) {
-    on.exit(expr = h5d$close())
+    on.exit(expr = h5d$close(), add = TRUE)
   }
   if (!inherits(x = h5d, what = "H5D")) {
     stop("\n  '", name, "' is not an H5D")
@@ -546,7 +560,7 @@
   if (!is.null(x = name)) {
     h5obj <- h5Open(x = h5group, name = name)
     if (!identical(x = h5obj, y = h5group)) {
-      on.exit(expr = h5obj$close())
+      on.exit(expr = h5obj$close(), add = TRUE)
     }
     if (inherits(x = h5obj, what = "H5D")) {
       return(h5ReadDataset(x = h5obj, transpose = transpose))
@@ -576,10 +590,10 @@
   if (!is.function(x = toS4.func)) {
     stop("\n  'toS4.func' must be NULL or a function.")
   }
-  return(toS4.func(r_obj))
+  S4Class <- .h5attr(h5obj = h5group, which = "S4Class")
+  return(toS4.func(r_obj, S4Class))
 }
 
-
 ## Assertive helpers ===========================================================
 #' @importFrom hdf5r H5D
 .assert_h5d <- function(h5obj) {
@@ -736,7 +750,9 @@
     ...
   )
   h5d <- h5Open(x = h5group, name = name)
-  on.exit(expr = h5d$close(), add = TRUE)
+  if (!identical(x = h5d, y = h5group)) {
+    on.exit(expr = h5d$close(), add = TRUE)
+  }
   idx_list <- NULL
   if (transpose) {
     idx_list <- lapply(X = dims, FUN = seq_len)
@@ -856,7 +872,7 @@
   if (ncol(x = robj) == 0) {
     ## Must add an empty 'column-order' to match anndata's IOSpec
     h5obj <- h5group$open(name = name)
-    on.exit(expr = h5obj$close())
+    on.exit(expr = h5obj$close(), add = TRUE)
     h5space <- H5S$new(dims = 0, maxdims = 0)
     on.exit(expr = h5space$close(), add = TRUE)
     h5a <- h5obj$create_attr(
@@ -932,6 +948,7 @@
   return(invisible(x = NULL))
 }
 
+#' @importClassesFrom Matrix dgCMatrix
 .h5_set_sparse_attrs <- function(
     robj,
     h5group, 
@@ -1023,7 +1040,6 @@
   return(invisible(x = NULL))
 }
 
-#' @importFrom hdf5r h5garbage_collect
 .h5read_factor <- function(h5obj) {
   h5codes <- h5obj$open(name = 'codes')
   on.exit(expr = h5codes$close(), add = TRUE)
@@ -1031,14 +1047,13 @@
   on.exit(expr = h5categories$close(), add = TRUE)
   codes <- h5codes$read()
   categories <- h5categories$read()
-  h5garbage_collect()
   r_obj <- factor(x = categories[codes + 1L], levels = categories)
   return(r_obj)
 }
 
 .h5read_list <- function(h5obj, transpose = FALSE, ...) {
-  r_obj <- list()
   elem_names <- h5obj$names
+  r_obj <- list()
   if (length(x = elem_names) == 0) {
     return(r_obj)
   }
@@ -1050,18 +1065,18 @@
 
 .h5get_column_order <- function(h5obj) {
   h5a <- h5obj$attr_open(attr_name = "column-order")
-  on.exit(expr = h5a$close())
+  on.exit(expr = h5a$close(), add = TRUE)
   h5space <- h5a$get_space()
   on.exit(expr = h5space$close(), add = TRUE)
   if (h5space$dims == 0) {
-    return(invisible(x = NULL))
+    return(NULL)
   }
   return(h5a$read())
 }
 
 .h5read_dataframe <- function(h5obj) {
   col_orders <- .h5get_column_order(h5obj = h5obj)
-  index <- h5Attr(x = h5obj, which = "_index")
+  index <- .h5attr(h5obj = h5obj, which = "_index")
   r_list <- .h5read_list(h5obj = h5obj)
   rownames <- NULL
   if (!is.null(x = index)) {
@@ -1078,17 +1093,16 @@
 }
 
 #' @importFrom Matrix sparseMatrix
-#' @importFrom hdf5r h5garbage_collect
 #' @importMethodsFrom Matrix t
 .h5read_sparse <- function(h5obj, transpose = FALSE) {
-  dims <- h5Attr(x = h5obj, which = "shape")
+  dims <- .h5attr(h5obj = h5obj, which = "shape")
   h5data <- h5obj$open(name = "data")
   on.exit(expr = h5data$close(), add = TRUE)
   h5indices <- h5obj$open(name = "indices")
   on.exit(expr = h5indices$close(), add = TRUE)
   h5indptr <- h5obj$open(name = "indptr")
   on.exit(expr = h5indptr$close(), add = TRUE)
-  encode <- h5Attr(x = h5obj, which = "encoding-type")
+  encode <- .h5attr(h5obj = h5obj, which = "encoding-type")
   m <- switch(
     EXPR = encode,
     "csr_matrix" = sparseMatrix(
@@ -1108,7 +1122,6 @@
       repr = "R"
     )
   )
-  h5garbage_collect()
   if (h5Exists(x = h5obj, name = "row_names")) {
     rownames(x = m) <- h5ReadDataset(x = h5obj, name = "row_names")
   }
